/*
 * Copyright (C) 2013 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Pawel Stolowski <pawel.stolowski@canonical.com>
 *
 */

namespace Unity.HomeScope
{
  /**
    * Maintains master scopes instances.
    */
  public class MasterScopesManager
  {
    private struct CategoryDefinition
    {
      string id;
      string name;
      string icon_hint;
      string sort_field;
      AggregatorScope.SortFlags sort_flags;
      string dedup_field;
    }

    private struct MasterDefinition
    {
      CategoryDefinition[] categories;
      FilterSet filters;
    }

    const string ICON_PATH = Config.DATADIR + "/icons/unity-icon-theme/places/svg/";
    private List<Unity.DeprecatedScopeBase?> master_scopes = new List<Unity.DeprecatedScopeBase?> ();

    public void start ()
    {
      HashTable<string, MasterDefinition?> defs = new HashTable<string, MasterDefinition?> (str_hash, str_equal);

      var filterset = new Unity.FilterSet ();

      defs["more_suggestions.scope"] = MasterDefinition () {
        categories = { CategoryDefinition () { id = "more_suggestions", name = _("More suggestions"), icon_hint = null, sort_field = "uri" } }
      };

      defs["reference.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "encyclopedia", name = _("Encyclopedia"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "vocabulary", name = _("Vocabulary"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "scholar", name = _("Scholar"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "sources", name = _("Sources"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["info.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "results", name = _("Results"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "weather", name = _("Weather"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "locations", name = _("Locations"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["news.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "recent", name = _("Recent"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "top", name = _("Top"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "local", name = _("Local"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["graphics.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "top", name = _("Top"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "recent", name = _("Recent"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["notes.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "recent", name = _("Recent"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "most-used", name = _("Most used"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["help.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "help", name = _("Help"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "community", name = _("Community"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["code.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "documentation", name = _("Documentation"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "code", name = _("Code"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "people", name = _("People"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["recipes.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "recipes", name = _("Recipes"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["calendar.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "upcoming", name = _("Upcoming"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "recent", name = _("Recent"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["web.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "history", name = _("History"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "bookmarks", name = _("Bookmarks"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["books.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "recent", name = _("Recent"), icon_hint = null, dedup_field = "uri" },
          CategoryDefinition () { id = "top", name = _("Top"), icon_hint = null, dedup_field = "uri" }
        }
      };

      defs["boxes.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "boxes", name = _("Boxes"), icon_hint = null, dedup_field = "uri" }
        }
      };

      filterset = new FilterSet ();
      {
        var filter = new RadioOptionFilter ("modified", _("Last modified"));

        filter.add_option ("last-7-days", _("Last 7 days"));
        filter.add_option ("last-30-days", _("Last 30 days"));
        filter.add_option ("last-year", _("Last year"));

        filterset.add (filter);
      }
      {
        var filter = new CheckOptionFilter ("type", _("Type"));
        filter.sort_type = OptionsFilter.SortType.DISPLAY_NAME;

        filter.add_option ("documents", _("Documents"));
        filter.add_option ("folders", _("Folders"));
        filter.add_option ("images", _("Images"));
        filter.add_option ("audio", _("Audio"));
        filter.add_option ("videos", _("Videos"));
        filter.add_option ("presentations", _("Presentations"));
        filter.add_option ("other", _("Other"));

        filterset.add (filter);
      }
      {
        var filter = new MultiRangeFilter ("size", _("Size"));

        filter.add_option ("1KB", _("1KB"));
        filter.add_option ("100KB", _("100KB"));
        filter.add_option ("1MB", _("1MB"));
        filter.add_option ("10MB", _("10MB"));
        filter.add_option ("100MB", _("100MB"));
        filter.add_option ("1GB", _("1GB"));
        filter.add_option (">1GB", _(">1GB"));

        filterset.add (filter);
      }

      defs["files.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "recent", name = _("Recent"), icon_hint = "group-recent.svg", dedup_field = "uri" },
          CategoryDefinition () { id = "downloads", name = _("Downloads"), icon_hint = "group-downloads.svg", dedup_field = "uri" },
          CategoryDefinition () { id = "folders", name = _("Folders"), icon_hint = "group-folders.svg", dedup_field = "uri" }
        },
        filters = filterset
      };

      defs["video.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "local", name = _("My Videos"), icon_hint = "group-videos.svg", dedup_field = "uri" },
          CategoryDefinition () { id = "online", name = _("Online"), icon_hint = "group-internet.svg", dedup_field = "uri" },
          CategoryDefinition () { id = "more", name = _("More suggestions"), icon_hint = "group-treat-yourself.svg", dedup_field = "uri" },
        }
      };

      filterset = new FilterSet ();
      {
        var filter = new MultiRangeFilter ("decade", _("Decade"));

        filter.add_option ("0", _("Old"));
        filter.add_option ("1960", _("60s"));
        filter.add_option ("1970", _("70s"));
        filter.add_option ("1980", _("80s"));
        filter.add_option ("1990", _("90s"));
        filter.add_option ("2000", _("00s"));
        filter.add_option ("2010", _("10s"));

        filterset.add (filter);
      }
      {
        var filter = new CheckOptionFilterCompact ("genre", _("Genre"));
        filter.sort_type = OptionsFilter.SortType.DISPLAY_NAME;

        filter.add_option ("blues", _("Blues"));
        filter.add_option ("classical", _("Classical"));
        filter.add_option ("country", _("Country"));
        filter.add_option ("disco", _("Disco"));
        filter.add_option ("funk", _("Funk"));
        filter.add_option ("rock", _("Rock"));
        filter.add_option ("metal", _("Metal"));
        filter.add_option ("hip-hop", _("Hip-hop"));
        filter.add_option ("house", _("House"));
        filter.add_option ("new-wave", _("New-wave"));
        filter.add_option ("r-and-b", _("R&B"));
        filter.add_option ("punk", _("Punk"));
        filter.add_option ("jazz", _("Jazz"));
        filter.add_option ("pop", _("Pop"));
        filter.add_option ("reggae", _("Reggae"));
        filter.add_option ("soul", _("Soul"));
        filter.add_option ("techno", _("Techno"));
        filter.add_option ("other", _("Other"));

        filterset.add (filter);
      }

      defs["music.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "songs", name = _("Songs"), icon_hint = "group-songs.svg", dedup_field = "uri" },
          CategoryDefinition () { id = "albums", name = _("Albums"), icon_hint = "group-albums.svg", dedup_field = "uri" },
          CategoryDefinition () { id = "radio", name = _("Radio"), icon_hint = "group-songs.svg", dedup_field = "uri" },
          CategoryDefinition () { id = "more", name = _("More suggestions"), icon_hint = "group-treat-yourself.svg", dedup_field = "uri" },
        },
        filters = filterset
      };

      filterset = new FilterSet ();
      {
        var filter = new RadioOptionFilter ("date", _("Date"));

        filter.add_option ("7", _("Last 7 days"));
        filter.add_option ("30", _("Last 30 days"));
        filter.add_option ("180", _("Last 6 months"));
        filter.add_option ("100000", _("Older"));

        filterset.add (filter);
      }

      defs["photos.scope"] = MasterDefinition () {
        categories = {
          CategoryDefinition () { id = "recent", name = _("Recent"), icon_hint = "group-recent.svg", dedup_field = "uri", sort_field = "comment", sort_flags = AggregatorScope.SortFlags.DESCENDING},
          CategoryDefinition () { id = "mine", name = _("My Photos"), icon_hint = "group-photos.svg", dedup_field = "uri", sort_field = "comment", sort_flags = AggregatorScope.SortFlags.DESCENDING},
          CategoryDefinition () { id = "friends", name = _("Friends"), icon_hint = "group-friends.svg", dedup_field = "uri", sort_field = "comment", sort_flags = AggregatorScope.SortFlags.DESCENDING},
          CategoryDefinition () { id = "online", name = _("Online Photos"), icon_hint = "group-internet.svg", dedup_field = "uri", sort_field = "comment", sort_flags = AggregatorScope.SortFlags.DESCENDING}
        },
        filters = filterset
      };

      var icon_dir = File.new_for_path (ICON_PATH);

      foreach (var node in ScopeRegistry.instance ().scopes)
      {
        if (node.scope_info.is_master)
        {
          var info = node.scope_info;
          var master_def = defs.lookup (info.id);
          if (master_def != null)
          {
            var scope = new Unity.MasterScope (info.dbus_path, info.id);
            scope.search_hint = info.search_hint;
            var category_set = new Unity.CategorySet ();

            Icon? main_icon = null;
            if (info.category_icon != null && info.category_icon.length > 0)
            {
              var icon_file = File.new_for_path (info.category_icon);
              main_icon = new GLib.FileIcon (icon_file);
            }
            else
            {
              main_icon = new GLib.ThemedIcon ("generic"); //fallback
            }

            // all masters need to have 'global' category
            category_set.add (new Unity.Category ("global", info.name, main_icon));

            foreach (var cat_def in master_def.categories)
            {
              Icon? icon = null;
              if (cat_def.icon_hint != null)
                icon = new FileIcon (icon_dir.get_child (cat_def.icon_hint));
              else
                icon = main_icon;

              var cat = new Unity.Category (cat_def.id, cat_def.name, icon);
              category_set.add (cat);
            }
            scope.categories = category_set;
            if (master_def.filters != null)
            {
              scope.filters = master_def.filters;
            }

            for (int cat_index = 0; cat_index < master_def.categories.length; cat_index++)
            {
              var cat_def = master_def.categories[cat_index];
              string dedup_field = cat_def.dedup_field;
              if (dedup_field != null && dedup_field != "")
                scope.add_constraint (cat_index, dedup_field);

              string sort_field = cat_def.sort_field;
              if (sort_field != null && sort_field != "")
                scope.add_sorter (cat_index, sort_field, cat_def.sort_flags);
              
              ++cat_index;
            }
            master_scopes.append (scope);
          
            debug ("Start master scope: %s", node.scope_info.id);
              
            try
            {
              scope.export ();
            }
            catch (Error e)
            {
              warning ("Failed to export master scope %s: %s", node.scope_info.id, e.message);
            }
          } // else we don't know about this master - it can be an external master scope
        }
      }
    }
  }
}
