/*
 * Copyright (C) 2013 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Pawel Stolowski <pawel.stolowski@canonical.com>
 *
 */

namespace Unity.HomeScope {

const string ICON_PATH = Config.DATADIR + "/icons/unity-icon-theme/places/svg/";
const int ICON_COLUMN = 1;

// metrics will get flushed and sent to the server after METRICS_MIN_NUM_EVENTS, or every METRICS_SEND_INTERVAL (whichever comes first).
const int METRICS_SEND_INTERVAL_SECS = 600;
const int METRICS_MIN_NUM_EVENTS = 10;
const int REMOTE_SCOPES_INITIAL_RETRY_INTERVAL_SECS = 2; //initial one, it will get increased with each retry

public class HomeScope : Unity.AggregatorScope
{
  private ScopeManager scope_mgr = new ScopeManager ();
  private SearchQueryState query_state = new SearchQueryState ();
  private KeywordSearch keywords_search = new KeywordSearch ();
  private bool smart_scopes_initialized = false;
  private bool smart_scopes_ready = false;
  private SmartScopes.SmartScopeClientInterface sss_client = null;
  private SmartScopes.ChannelIdMap channel_id_map = new SmartScopes.ChannelIdMap ();
  private uint metrics_timer;
  private int remote_scopes_retry_count = 0;
  private RemoteScopeRegistry remote_scope_registry;

  public HomeScope ()
  {
    Object (dbus_path: "/com/canonical/unity/home",
            id:"home.scope",
            merge_mode: Unity.AggregatorScope.MergeMode.OWNER_SCOPE);
  }

  protected override void constructed ()
  {
    base.constructed ();
    discover_scopes_sync ();

    CategoryManager.instance ().set_default_sort_order_from_dconf ();

    var reg = ScopeRegistry.instance ();
    if (reg.scopes == null)
    {
      critical ("No scopes found. Please check your installation");
      return;
    }

    search_hint = _("Search your computer and online sources");

    keywords_search.rebuild ();

    populate_filters ();
    populate_categories ();

    debug ("Starting master scopes");
    scope_mgr.start_master_scopes ();

    debug ("Exporting home scope");
    export ();

    init_sss_client ();
  }

  private static bool discovery_started = false;

  public static void discover_scopes_sync ()
  {
    if (discovery_started) return;
    discovery_started = true;

    debug ("Starting scope discovery");

    var ml = new MainLoop ();

    var reg = ScopeRegistry.instance ();
    reg.find_scopes.begin (Config.PKGDATADIR + "/scopes", (obj, res) =>
    {
      reg.find_scopes.end (res);
      ml.quit ();
    });
    ml.run ();
  }

  internal void init_sss_client ()
  {
    if (!smart_scopes_initialized)
    {
      debug ("Initializing Smart Scopes client");

      ClientScopesInfo client_scopes_info = null;

      try
      {
        client_scopes_info = ClientScopesInfo.from_file (ClientScopesInfo.CLIENT_SCOPES_FILE, ScopeRegistry.instance ().flatten ());
      }
      catch (Error e)
      {
        warning ("Cannot determine client scopes: %s", e.message);
      }

      var platform_info = SmartScopes.PlatformInfo.gather_platform_info (client_scopes_info);

      smart_scopes_initialized = true;
      sss_client = new SmartScopes.SmartScopesClient (platform_info);

      request_remote_scopes.begin ();
      
      metrics_timer = GLib.Timeout.add_seconds (METRICS_SEND_INTERVAL_SECS, flush_metrics_async);
    }
  }

  internal async void request_remote_scopes ()
  {
    SmartScopes.RemoteScopeInfo[]? scopes = null;
    // issue a request for remote_scopes list
    try
    {
      scopes = yield sss_client.remote_scopes (null);
      if (scopes != null)
      {
        debug ("Got %u remote scopes", scopes.length);
        smart_scopes_ready = true;

        remote_scope_registry = new RemoteScopeRegistry.for_scopes (scopes);
        remote_scope_registry.create_model ();
      }
    }
    catch (Error e)
    {
      warning ("Failed to get list of remote scopes: %s", e.message);
    }

    if (!smart_scopes_ready)
    {
      // setup a retry callback
      GLib.Timeout.add_seconds (REMOTE_SCOPES_INITIAL_RETRY_INTERVAL_SECS * ++remote_scopes_retry_count, request_remote_scopes_retry);
    }
  }

  internal bool request_remote_scopes_retry ()
  {
    request_remote_scopes ();
    return false; // don't run this timeout callback anymore
  }

  internal void populate_filters ()
  {
    var filter_list = new FilterSet ();

    var cat_filter = new Unity.CheckOptionFilter ("categories", _("Categories"));
    cat_filter.show_all_button = false;
    cat_filter.sort_type = Unity.OptionsFilter.SortType.DISPLAY_NAME;

    var src_filter = new Unity.CheckOptionFilter ("sources", _("Sources"));
    foreach (var node in ScopeRegistry.instance ().scopes)
    {
      debug ("Category filter add: %s", node.scope_info.id);
      cat_filter.add_option (node.scope_info.id, node.scope_info.name); //add master scope id to 'Categories'
      foreach (var scope in node.sub_scopes)
      {
        debug ("Sources filter add: %s", scope.id);
        src_filter.add_option (scope.id, scope.name);
      }
    }
    src_filter.show_all_button = false;
    filter_list.add (cat_filter);
    filter_list.add (src_filter);

    filters = filter_list;
  }

  /**
    * Populate home scope category list with initial categories,
    * i.e. master scopes that have 'always-search' on.
    */
  internal void populate_categories ()
  {
    debug ("Setting home scope categorties");

    var cats = new CategorySet ();

    foreach (var node in ScopeRegistry.instance ().scopes)
    {
      var scope_info = node.scope_info;
      var scope_id = scope_info.id;

      debug ("Adding home scope category: %s", scope_id);
      CategoryManager.instance ().register (scope_id);
      var icon = new GLib.FileIcon (File.new_for_path (scope_info.category_icon ?? scope_info.icon));
      var cat = new Unity.Category (scope_id, scope_info.name, icon);
      cats.add (cat);
    }
    categories = cats; //set scope categories
  }

  // TODO: hook it up once API support it
  private void on_home_channel_closed (string channel_id)
  {
    query_state.remove (channel_id);
    channel_id_map.remove_channel (channel_id);
  }

  /**
    * Creates new session id if search string changed considerably, or returns existing session id
    */
  private string get_session_id (SearchQueryChange query_status, string home_channel_id)
  {
    string session_id = null;
    if (query_status != SearchQueryChange.NEW_QUERY)
    {
      if (!channel_id_map.has_session_id_for_channel (home_channel_id))
        warning ("no smart scopes session mapping for channel %s", home_channel_id);
      else
        session_id = channel_id_map.session_id_for_channel (home_channel_id);
    }
    if (session_id == null)
    {
      session_id = sss_client.create_session_id ();
      channel_id_map.map_session_id (home_channel_id, session_id);
    }
    return session_id;
  }

  internal bool flush_metrics_async ()
  {
    handle_feedback_sending.begin (true, (obj, res) =>
    {
      handle_feedback_sending.end (res);
    });
    return true;
  }

  internal async void handle_metrics (Unity.AggregatorActivation activation, string session_id, string server_sid)
  {
    var timestamp = new DateTime.now_utc ();

    if (activation.action_type == Unity.Protocol.ActionType.ACTIVATE_RESULT)
    {
      sss_client.add_click_event (session_id, server_sid, activation.scope_id, timestamp);
    }
    else if (activation.action_type == Unity.Protocol.ActionType.PREVIEW_RESULT)
    {
      sss_client.add_preview_event (session_id, server_sid, activation.scope_id, timestamp);
    }

    yield handle_feedback_sending ();
  }

  internal async void handle_found_metrics (string home_channel_id, string session_id, string server_sid, List<SmartScopes.RecommendedScope?> recommendations)
  {
    var cat_mgr = CategoryManager.instance ();

    var result_counts = new Gee.HashMap<string, int> ();
    foreach (var scope_rec in recommendations)
    {
      if (scope_rec.scope_type == SmartScopes.ScopeType.ClientScope) //only send result counts for client scope
      {
        int count = cat_mgr.get_result_count (home_channel_id, scope_rec.scope_id);
        if (count >= 0)
          result_counts[scope_rec.scope_id] = count;
        else
          warning ("No result count for recommended scope %s, home channel %s", scope_rec.scope_id, home_channel_id);
      }
    }

    var timestamp = new DateTime.now_utc ();
    sss_client.add_found_event (session_id, server_sid, result_counts, timestamp);

    yield handle_feedback_sending ();
  }

  internal async void handle_feedback_sending (bool force = false)
  {
    if (force || sss_client.num_feedback_events () > METRICS_MIN_NUM_EVENTS)
    {
      try
      {
        yield sss_client.send_feedback (null); //TODO cancellable
      }
      catch (Error e)
      {
        warning ("Sending feedback failed: %s", e.message);
      }
    }
  }

  internal async ActivationResponse? handle_preview (Unity.AggregatorActivation activation, string session_id, string server_sid) throws Error
  {
    if (activation.scope_result == null)
    {
      warning ("activation on channel %s has no result", activation.channel_id);
      return null;
    }
    
    var content = activation.scope_result.metadata.lookup ("content"); //master scopes put 'metadata' in 'content'
    if (content == null)
    {
      warning ("'content' element not found"); //this shouldn't really happen
      return null;
    }

    var result_id = content.lookup_value ("id", VariantType.STRING);
    if (result_id == null)
    {
      warning ("result_id not present for activation request on channel %s, uri %s", activation.channel_id, activation.scope_result.uri);
      return null;
    }
    
    var preview = yield sss_client.preview (server_sid, session_id, result_id.get_string (), activation.scope_result, null); //TODO cancellable?
    if (preview == null)
    {
      warning ("Request for server preview for uri %s returned null", activation.scope_result.uri);
      return null;
    }
    
    return new ActivationResponse.with_preview (preview);
  }

  internal override async ActivationResponse? activate (Unity.AggregatorActivation activation)
  {
    debug ("Activation request for scope %s, action_type=%u", activation.scope_id, activation.action_type);
 
    string? server_sid = channel_id_map.server_sid_for_channel (activation.channel_id);
    string? session_id = channel_id_map.session_id_for_channel (activation.channel_id);

    if (sss_client != null)
      yield handle_metrics (activation, session_id, server_sid);

    // do nothing for local scopes - activation will be handled by actual scope
    if (sss_client == null || scope_mgr.is_server_scope (activation.scope_id) == false)
      return null;

    if (activation.action_type == Unity.Protocol.ActionType.PREVIEW_RESULT)
    {
      try
      {
        var preview = yield handle_preview (activation, session_id, server_sid);
        return preview;
      }
      catch (Error e)
      {
        warning ("Preview request for scope %s failed: %s", activation.scope_id, e.message);
      }
    }

    return null; // activation will be handled by actual scope
  }

  public override async void search (Unity.AggregatedScopeSearch scope_search)
  {
    debug ("Search query: %s", scope_search.search_string);

    init_sss_client ();

    CategoryManager.instance ().observe (scope_search.channel_id, scope_search.results_model);

    bool wait_for_sss_query_finished = false;
    uint num_scopes = 0;

    AsyncReadyCallback cb = (obj, res) =>
    {
      var search_obj = obj as Unity.AggregatedScopeSearch;
      try
      {
        var reply_dict = search_obj.search_scope.end (res);
        var iter = HashTableIter<string, Variant> (reply_dict);
        unowned string key;
        unowned Variant variant;

        while (iter.next (out key, out variant))
          scope_search.set_reply_hint (key, variant);
      }
      catch (Error err)
      {
        warning ("Unable to search scope: %s", err.message);
      }
      if (--num_scopes == 0 && wait_for_sss_query_finished == false)
        search.callback ();
    };

    // data models that are populated by Smart Scopes search and pushed to respective scopes
    var push_data = new HashTable<string, Dee.SerializableModel> (str_hash, str_equal); //scope_id -> model

    // compare current search string with previous one to see if it's filters change only;
    // this impacts how we treat online recommendations and filters.
    var search_query_changed = query_state.search_query_changed (scope_search.channel_id, scope_search.search_string);

    // ids of scopes recommended by Smart Scope Service
    var recommended_search_scopes = new List<SmartScopes.RecommendedScope?> ();

    // maps ids of scopes to search (master scopes or top-level scopes) to sub-scopes (according to filter or direct keyword search);
    // may map to null, in that case there is no filtering based on sub-scopes (all subscopes of given master will be queried).
    var search_scopes = new HashTable<string, Gee.Set<string>?> (str_hash, str_equal);

    // apply user filters only if search string is unchanged (or smart scopes are disabled completly).
    // this mean user set filters *after* entering a query and we apply them;
    // otherwise smart scopes recommendations will take precedence.
    if (search_query_changed == SearchQueryChange.NOT_CHANGED || sss_client == null)
      SearchUtil.scopes_to_query_from_filters (scope_search.get_filter ("sources") as Unity.OptionsFilter,
                                               scope_search.get_filter ("categories") as Unity.OptionsFilter,
                                               search_scopes);

    // initiate Smart Scopes Search (if enabled and query not empty)
    if (smart_scopes_ready == true)
    {
      if (search_query_changed == SearchQueryChange.EMPTY_QUERY)
      { 
        if (channel_id_map.has_session_id_for_channel (scope_search.channel_id))
        {
          debug ("Empty search, removing session mapping for channel %s", scope_search.channel_id);
          channel_id_map.remove_channel (scope_search.channel_id);
        }
      }
      else
      {
        var session_id = get_session_id (search_query_changed, scope_search.channel_id);
        wait_for_sss_query_finished = true;

        AsyncReadyCallback sss_cb = (obj, res) =>
        {
          try
          {
            sss_client.search.end (res);
          }
          catch (Error e)
          {
            warning ("Remote search failed: %s", e.message);
          }
          wait_for_sss_query_finished = false;
          if (num_scopes == 0)
            search.callback ();
        };

        // apply filter to remote scopes query; since we don't really know remote scopes,
        // we can only pass master scopes ids.
        
        var reg = ScopeRegistry.instance ();
        string[] remote_scopes_to_query = new string [0];
        foreach (var scope_id in search_scopes.get_keys ())
        {
          if (reg.is_master (scope_id)) //make sure it's really a master, we don't want 'applications.scope' etc.
            remote_scopes_to_query += scope_id;
        }

        sss_client.search.begin (scope_search.search_string, session_id, remote_scopes_to_query,
            (scope_id, row) =>
            {
              var pushed_model = push_data.lookup (scope_id);
              if (pushed_model == null)
              {
                pushed_model = new Dee.SequenceModel ();
                pushed_model.set_schema_full ({"s", "s", "u", "u", "s", "s", "s", "s", "a{sv}"});
                push_data[scope_id] = pushed_model;
              }
              pushed_model.append_row (row);
            },
            (scope_id, categories) =>
            {
              //TODO handle categories
            },
            (server_sid, recommendations) =>
            {
              if (server_sid != null)
              {             
                channel_id_map.map_server_sid (scope_search.channel_id, server_sid);
                foreach (var scope_rec in recommendations)
                  recommended_search_scopes.append (scope_rec);
              }
              else
              {
                warning ("server_sid is null");
              }
            }, null, //TODO cancellable
            sss_cb);
      }
    }
    else
    {
      // may happen if remote-scopes haven't finished or returned an error
      warning ("Smart scopes not ready and not enabled for this query");
    }

    // no filters set / no scopes to search, use always-search scopes
    if (search_scopes.size () == 0)
      SearchUtil.scopes_to_query_from_requested_ids (scope_mgr.get_always_run_scopes (), search_scopes);

    // handle keywords
    string search_string;
    bool direct_search = false;
    unowned Gee.Set<string>? requested_scope_ids = keywords_search.process_query (scope_search.search_string, out search_string);
    if (requested_scope_ids != null && search_string != null)
    {
      debug ("Potential keyword-based query, search_string = '%s'", search_string);
      direct_search = SearchUtil.scopes_to_query_from_requested_ids (requested_scope_ids, search_scopes);
    }

    num_scopes = search_scopes.size ();
    debug ("Dispatching search to %u scopes, home_channel=%s", num_scopes, scope_search.channel_id);

    if (search_string == null)
      search_string = scope_search.search_string;

    // iterate over master scopes, dispatch search query
    foreach (var scope_id in search_scopes.get_keys ())
    {
      debug ("Dispatching search to scope %s", scope_id);
      // apply Sources (subscopes) filtering; pass subscopes requested via filters to relevant masterscopes
      var hints = new HashTable<string, Variant> (str_hash, str_equal);
      SearchUtil.set_subscopes_filter_hint (hints, search_scopes, scope_id);
      scope_search.search_scope.begin (scope_id, search_string,
                                        SearchType.GLOBAL, hints, cb);
    }

    bool use_recommended_scopes = false;
    
    // filters are updated during the search, but may need to be updated again afterwards if we find out there wero no results for some categories
    bool needs_filter_update = false;

    // dispatch search to scopes recommended by smart scopes service
    if (smart_scopes_ready  && search_query_changed != SearchQueryChange.EMPTY_QUERY)
    {
      // wait for smart scopes service query to finish
      if (wait_for_sss_query_finished == true)
        yield;

      foreach (var scope_id in push_data.get_keys ())
      {
        message ("Pushing results (%u rows) for scope %s", push_data[scope_id].get_n_rows (), scope_id);
        yield scope_search.push_results (scope_id, push_data[scope_id], {"global"});
      }

      debug ("Got %u recommended scopes from Smart Scope Service", recommended_search_scopes.length ());
    }

    // only use recommended scopes if search query was changed (thus user-selected filters are reset).
    if (search_query_changed != SearchQueryChange.NOT_CHANGED && !direct_search && recommended_search_scopes.length () > 0)
      use_recommended_scopes = true;

    debug ("Updating filter state");
    var filters = SearchUtil.update_filters (search_scopes, use_recommended_scopes ? recommended_search_scopes : null, scope_search, false, out needs_filter_update);
    if (filters != null)
    {
      debug ("Sending updated filters");
      scope_search.push_filter_settings (filters);
    }
    
    if (use_recommended_scopes)
    {
        debug ("Search query changed, querying recommended scopes");

        // ids of recommended scopes
        var extra_search_scopes = new HashTable<string, Gee.Set<string>?> (str_hash, str_equal);

        // build the list of master-scopes to search and subscopes_filter to pass to them based on recommendations.
        // note that recommendations may contains subscopes, so build_search_scopes_list will take care of all the magic.
        // also note that, because of that, we can't filter out master scopes that were already searched before (as part
        // of always-search list) just now - this happens below in the search loop.
        foreach (var scope_rec in recommended_search_scopes)
        {
          if (scope_rec.scope_type == SmartScopes.ScopeType.ClientScope)
            SearchUtil.build_search_scopes_list (scope_rec.scope_id, extra_search_scopes);
        }

        foreach (var scope_id in extra_search_scopes.get_keys ())
        {
          // ignore this master if it was already searched before.
          if (!search_scopes.contains (scope_id))
          {
            debug ("Dispatching search to recommended scope %s", scope_id);
            ++num_scopes;
            var hints = new HashTable<string, Variant> (str_hash, str_equal);
            SearchUtil.set_subscopes_filter_hint (hints, extra_search_scopes, scope_id);
            scope_search.search_scope.begin (scope_id, search_string, SearchType.GLOBAL, hints, cb);
          }
        }
    }
    else
    {
      debug ("Search query not changed or direct search active, ignoring recommended scopes");
    }

    // wait for the results from recommended scopes
    if (num_scopes > 0)
      yield;
    
    debug ("search finished");
    
    //FIXME: make sure we're synchronized after pushing data

    signal_categories_order (scope_search, recommended_search_scopes);

    // update filter state again, but this time check result counts and send the update only if any of the highlighted masters has no results
    filters = SearchUtil.update_filters (search_scopes, use_recommended_scopes ? recommended_search_scopes : null, scope_search, true, out needs_filter_update);
    if (filters != null && needs_filter_update)
    {
      debug ("Sending updated filters");
      scope_search.push_filter_settings (filters);
    }

    if (use_recommended_scopes)
    {
      debug ("Updating filter state with recommendations");

      string? server_sid = channel_id_map.server_sid_for_channel (scope_search.channel_id);
      string? session_id = channel_id_map.session_id_for_channel (scope_search.channel_id);

      yield handle_found_metrics (scope_search.channel_id, session_id, server_sid, recommended_search_scopes);
    }
  }

  public override int category_index_for_scope_id (string scope_id)
  {
    return CategoryManager.instance ().get_category_index (scope_id);
  }
  
  private void signal_categories_order (AggregatedScopeSearch search, List<SmartScopes.RecommendedScope?> recommended_scopes)
  {
    var cats = CategoryManager.instance ().get_category_order (search.search_string, search.channel_id, search.results_model, recommended_scopes);
    debug ("Updating categories order");
    search.category_order_changed (cats); //TODO only signal if order really changed?
  }
}

}
