/*
 * Copyright (C) 2013 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Pawel Stolowski <pawel.stolowski@canonical.com>
 *
 */

namespace Unity.HomeScope.SearchUtil
{
  public Unity.FilterSet? update_filters (HashTable<string, Gee.Set<string>?> search_scopes,
                                                               List<SmartScopes.RecommendedScope?>? recommendations,
                                                               Unity.AggregatedScopeSearch scope_search,
                                                               bool check_result_counts, out bool has_zero_results)
  {
    // will be set to true if check_result_counts is true and any of the recommended/searched scopes has 0 results
    has_zero_results = false;

    // create a lookup for recommended scopes ids and search_scopes.
    // recommendations include only specific scopes, so for every recommended scope insert its master as well.
    var rec_lookup = new Gee.TreeSet<string> ();

    if (recommendations != null)
    {
      foreach (var scope in recommendations)
      {
        var parts = scope.scope_id.split ("-", 2);
        string master_scope_id = "%s.scope".printf (parts[0]);
        
        rec_lookup.add (scope.scope_id);
        rec_lookup.add (master_scope_id);
      }
    }

    search_scopes.foreach ((master, subscopes) =>
    {
      rec_lookup.add (master);
      if (subscopes != null)
      {
        foreach (var scope in subscopes)
          rec_lookup.add (scope);
      }
    });

    var home_channel = scope_search.channel_id;
    var categories = scope_search.get_filter ("categories") as Unity.OptionsFilter;
    var sources = scope_search.get_filter ("sources") as Unity.OptionsFilter;
    var mgr = CategoryManager.instance ();

    if (categories != null)
    {
      foreach (var opt in categories.options) //master scopes
      {
        int result_count = mgr.get_result_count (home_channel, opt.id);
        debug ("Results for %s: %d", opt.id, result_count);
        if (rec_lookup.contains (opt.id))
        {
          opt.active = check_result_counts == false || result_count > 0;
          if (result_count == 0 && check_result_counts)
            has_zero_results = true;
        }
        else
        {
          opt.active = false;
        }
        debug ("Setting category filter %s: %s", opt.id, opt.active.to_string ());
      }
    }
    else
    {
      warning ("Couldn't get categories filter");
    }

    if (sources != null)
    {
      foreach (var opt in sources.options)
      {
        var parts = opt.id.split ("-", 2);
        string master_scope_id = "%s.scope".printf (parts[0]);

        // we don't have result counts for specific scopes, so light them up if master has any results
        int result_count = mgr.get_result_count (home_channel, master_scope_id);
        if (rec_lookup.contains (opt.id))
        {
          opt.active = check_result_counts == false || result_count > 0;
          if (result_count == 0 && check_result_counts)
            has_zero_results = true;
        }
        else
        {
          opt.active = false;
        }
        debug ("Setting sources filter %s: %s", opt.id, opt.active.to_string ());
      }
    }
    else
    {
      warning ("Couldn't get sources filter");

      if (categories == null)
        return null; // no filters
    }

    var filters = new Unity.FilterSet ();
    if (categories != null)
      filters.add (categories);
    if (sources != null)
      filters.add (sources);

    return filters;
  }

  public void build_search_scopes_list (string scope_id, HashTable<string, Gee.Set<string>?> search_scopes)
  {
    var registry = ScopeRegistry.instance ();
    if (!registry.is_master (scope_id))
    {
      var master_id = registry.get_master_scope_id (scope_id);
      if (master_id != null) // if null, then treat it as a non-master scope that connects directly to home
      {
        var subscopes = search_scopes.lookup (master_id);
        if (subscopes == null)
        {
          subscopes = new Gee.TreeSet<string?> ();
          search_scopes[master_id] = subscopes;
        }
        subscopes.add (scope_id);
        return;
      }
    }
    search_scopes[scope_id] = null;
  }

  /**
   * Populate search hints with "subscopes_filter" value, suitable for passing to search_scope.
   */
  public void set_subscopes_filter_hint (HashTable<string, Variant> hints, HashTable<string, Gee.Set<string>?> search_scopes, string scope_id)
  {
    var subscopes = search_scopes.lookup (scope_id);
    if (subscopes != null)
    {
      var src_filter_var = new GLib.Variant.strv (subscopes.to_array ());
      hints ["subscopes_filter"] = src_filter_var;
    }
  }

  /**
    Populate list of master scopes and their subscopes to query based on sources and categories filters.
    search_scopes maps master scope_id to its subscopes.
   */
  public void scopes_to_query_from_filters (OptionsFilter sources_filter, OptionsFilter cat_filter, HashTable<string, Gee.Set<string>?> search_scopes)
  {
    if (sources_filter == null)
      return;

    Gee.Set<string> requested_ids = new Gee.TreeSet<string> ();

    if (sources_filter.filtering) // subscopes-filter
    {
      foreach (var opt in sources_filter.options) //each option is a subscope id
      {
        if (opt.active)
          requested_ids.add (opt.id);
      }
    }

    if (cat_filter.filtering)
    {
      foreach (var cat in cat_filter.options)
      {
        if (cat.active)
          requested_ids.add (cat.id);
      }
    }

    scopes_to_query_from_requested_ids (requested_ids, search_scopes);
  }

  public bool scopes_to_query_from_requested_ids (Gee.Set<string> requested_scope_ids, HashTable<string, Gee.Set<string>?> search_scopes)
  {
    bool found = false;
    var registry = ScopeRegistry.instance ();

    foreach (var req_scope_id in requested_scope_ids)
    {
      if (registry.is_master (req_scope_id))
      {
        if (!found)
          search_scopes.remove_all ();

        if (!search_scopes.contains (req_scope_id))
          search_scopes[req_scope_id] = null;
        found = true;
      }
      else // user may have requested a subscope (not a master scope), we need to find a master that needs to receive the query
      {
        var master_id = registry.get_master_scope_id (req_scope_id);
        if (master_id != null)
        {
          if (!found)
            search_scopes.remove_all ();

          Gee.Set<string>? subscopes = search_scopes.lookup (master_id);
          if (subscopes == null)
          {
            subscopes = new Gee.TreeSet<string> ();
          }
          subscopes.add (req_scope_id);
          search_scopes[master_id] = subscopes;
          found = true;
        }
        else
        {
          warning ("Requested scope '%s' doesn't match any known scope id", req_scope_id);
        }
      }
    }
    return found;
  }
}
