/*
 * Copyright (C) 2013 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Pawel Stolowski <pawel.stolowski@canonical.com>
 *
 */

namespace Unity.HomeScope.SmartScopes {
  public class SmartScopesPreviewParser
  {
    private HashTable<string, PreviewJsonParser?> handlers = new HashTable<string, PreviewJsonParser?> (str_hash, str_equal);

    public SmartScopesPreviewParser ()
    {
      // preview-series is not used, so not supported here
      handlers["preview-generic"] = new PreviewGenericParser ();
      handlers["generic-preview"] = new PreviewGenericParser ();
      handlers["preview-music"] = new PreviewMusicParser ();
      handlers["preview-movie"] = new PreviewMovieParser ();
      handlers["preview-social"] = new PreviewSocialParser ();
    }

    public Preview? parse (Json.Object obj)
    {
      unowned string renderer_name = obj.get_string_member ("renderer_name");
      var parser = handlers.lookup (renderer_name);
      if (parser != null)
      {
        return parser.parse (obj);
      }

      warning ("No handler for preview type %s", renderer_name);
      return null;
    }
  }

  public interface PreviewJsonParser: Object
  {
    internal void get_base_attributes (Json.Object obj, out string title, out string subtitle, out string description, out GLib.Icon icon, out string attribution) //TODO actions
    {
      title = obj.get_string_member ("title");
      subtitle = obj.get_string_member ("subtitle");
      description = obj.get_string_member ("description");
      attribution = obj.get_string_member ("attribution");
      var icon_uri = obj.get_string_member ("image_hint");
      var icon_file = File.new_for_uri (icon_uri);
      icon = new GLib.FileIcon (icon_file);
    }

    public abstract Preview? parse (Json.Object obj);
  }

  public class PreviewGenericParser: PreviewJsonParser, Object
  {
    public Preview? parse (Json.Object obj)
    {
      string title, description, attribution;
      GLib.Icon? icon;

      get_base_attributes (obj, out title, null, out description, out icon, out attribution);
      var preview = new GenericPreview (title, description, icon);
      return preview;
    }
  }

  public class PreviewApplicationParser: PreviewJsonParser, Object
  {
    public Preview? parse (Json.Object obj)
    {
      string title, subtitle, description, attribution;
      GLib.Icon? icon;
      GLib.Icon? screenshot = null;

      get_base_attributes (obj, out title, out subtitle, out description, out icon, out attribution);

      var preview = new ApplicationPreview (title, subtitle, description, icon, screenshot);
      return preview;
    }
  }

  public class PreviewMusicParser: PreviewJsonParser, Object
  {
    public Preview? parse (Json.Object obj)
    {
      string title, subtitle, description, attribution;
      GLib.Icon? icon;

      get_base_attributes (obj, out title, out subtitle, out description, out icon, out attribution);

      var preview = new MusicPreview (title, subtitle, icon);
      return preview;
    }
  }

  public class PreviewMovieParser: PreviewJsonParser, Object
  {
    public Preview? parse (Json.Object obj)
    {
      string title, subtitle, description, attribution;
      GLib.Icon? icon;

      get_base_attributes (obj, out title, out subtitle, out description, out icon, out attribution);

      var preview = new MoviePreview (title, subtitle, description, icon);
      return preview;
    }
  }

  public class PreviewSocialParser: PreviewJsonParser, Object
  {
    public Preview? parse (Json.Object obj)
    {
      string title, subtitle, description, attribution;
      GLib.Icon? avatar = null;

      get_base_attributes (obj, out title, out subtitle, out description, null, out attribution);

      var preview = new SocialPreview (title, subtitle, description, avatar);
      return preview;
    }
  }
}
